#include <GL/glew.h>
#include <GL/freeglut.h>
#include <GL/glut.h>
#include <stdio.h>

// Súradnice vertexov a farby
static const GLfloat vertexBufferData[] = {
	-0.9f, -0.9f, 1, 0, 0,
	-0.9f,  0.9f, 0, 1, 0,
	 0.9f, -0.9f, 0, 0, 1,
	 0.9f,  0.9f, 1, 1, 1,
};
GLuint vertexBuffer;

// Kód vertex a fragment shadera
static const char vertexShader[] = "#version 330 core\n\
\n\
layout(location = 0) in vec2 position;\n\
layout(location = 1) in vec3 colorAttribute;\n\
\n\
out vec3 fragmentColor;\n\
\n\
void main(){\n\
	gl_Position = vec4(position, 0.0, 1.0);\n\
	fragmentColor = colorAttribute;\n\
}\n\
";
static const char fragmentShader[] = "#version 330 core\n\
in vec3 fragmentColor;\n\
\n\
out vec4 color;\n\
\n\
void main()\n\
{\n\
	color = vec4(fragmentColor, 1.0);\n\
}\n\
";
GLuint shaderProgram;

// Výpis chyby kompilácie ak shader nebol skompilovaný
void printShaderStatus(GLuint shader) {
	GLint res = GL_FALSE;
	int infoLength;

	glGetShaderiv(shader, GL_COMPILE_STATUS, &res);
	glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infoLength);
	if (res != GL_TRUE) {
		char *msg = malloc(infoLength + 1);
		glGetShaderInfoLog(shader, infoLength, NULL, msg);
		printf("%s", msg);
		free(msg);
	}
}

// Výpis chyby linkovania ak sa program nepodarilo zlinkovať
void printProgramStatus(GLuint program) {
	GLint res = GL_FALSE;
	int infoLength;

	glGetProgramiv(program, GL_LINK_STATUS, &res);
	glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infoLength);
	if (res != GL_TRUE) {
		char *msg = malloc(infoLength + 1);
		glGetProgramInfoLog(program, infoLength, NULL, msg);
		printf("%s", msg);
		free(msg);
	}
}

// Kompilácia ľubovoľného druhu shadera
void compileShader(GLuint shader, const char *code) {
	glShaderSource(shader, 1, &code, NULL);
	glCompileShader(shader);
	printShaderStatus(shader);
}

// Kompilácia shaderov a zlinkovanie do programu
GLuint getShaderProgram(const char *vertexShaderSource, const char *fragmentShaderSource) {
	// Vytvorenie programu
	GLuint program = glCreateProgram();
	// Vytvorenie inštancií shaderov
	GLuint vertexShader = glCreateShader(GL_VERTEX_SHADER);
	GLuint fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
	// Kompilácia oboch zdrojových kódov
	compileShader(vertexShader, vertexShaderSource);
	compileShader(fragmentShader, fragmentShaderSource);
	// Priradenie shaderov k jednému programu
	glAttachShader(program, vertexShader);
	glAttachShader(program, fragmentShader);
	// Zlinkovanie programu a výpis chyb ak nebolo linkovanie úspešné
	glLinkProgram(program);
	printProgramStatus(program);
	// Ďalej shader nnemusí byť priradený k programu
	glDetachShader(program, vertexShader);
	glDetachShader(program, fragmentShader);
	return program;
}


// Zobrazenie štvorca
void display() {
	// Vyčistenie pozadia
	glClear(GL_COLOR_BUFFER_BIT);

	// Pri vykresľovaní sa použije shader
	glUseProgram(shaderProgram);

	// Priradenie poľa k atribútu s location(0) (position)
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
	glVertexAttribPointer(
		0, // Index atribútu (zodpovedá hodnote v location)
		2, // Počet komponentov (zadávame pre každý bod iba 2 súradnice)
		GL_FLOAT, // V poli sú hodnoty s pohyblivou desatinnou čiarkou
		GL_FALSE, // Nemá význam pre typ float
		sizeof(GLfloat)*5, // Rozostup medzi prvkami
		(void *)0 // Posun voči začiatku poľa
	);

	// Priradenie poľa k atribútu s location(1) (colorAttribute)
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(GLfloat)*5, (void *)(sizeof(GLfloat)*2));

	// Vykreslenie objektu triangle strip zo 4 vertexov
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);

	// Zobrazenie na obrazovke
	glFlush();
}

/*
void GLAPIENTRY debugCallback(__attribute__((unused))GLenum source, GLenum type, __attribute__((unused))GLuint id, GLenum severity, __attribute__((unused))GLsizei length, const GLchar* message, __attribute__((unused))const void* userParam)
{
	fprintf(stderr, "GL: %s type = 0x%x, severity = 0x%x, message = %s\n", (type == GL_DEBUG_TYPE_ERROR ? "** ERROR **" : ""), type, severity, message);
}
*/

// Inicializácia OpenGL a hlavnej slučky
int main(int argc, char** argv) {
	glutInit(&argc, argv);

	// Explicitne vyžadujem OpenGL 3.3 core profile
	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);

	// Režim RGBA s double bufferom
	glutInitDisplayMode(GLUT_SINGLE | GLUT_RGBA);

	// Vytvorenie okna
	glutInitWindowSize(400, 300);
	glutCreateWindow("Okno");

	// Inicializácia GLEW musí nasledovať po otvorení okna
	glewInit();

	//glEnable(GL_DEBUG_OUTPUT);
	//glDebugMessageCallback(debugCallback, 0);

	// Aktivovaný vertex array objekt
	GLuint vertexArray;
	glGenVertexArrays(1, &vertexArray);
	glBindVertexArray(vertexArray);

	// Kompilácia
	shaderProgram = getShaderProgram(vertexShader, fragmentShader);

	// Upload vertexov a farieb do GPU
	glGenBuffers(1, &vertexBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertexBufferData), vertexBufferData, GL_STATIC_DRAW);

	// Zaregistrovanie funkcie, ktorá vykreslí obsah okna
	glutDisplayFunc(display);

	// Hlavná slučka
	glutMainLoop();
}
